@@ -1,5 +1,5 @@
 /*
-   Copyright (C) 2000-2008  Ulric Eriksson <ulric@siag.nu>
+   Copyright (C) 2000-2013  Ulric Eriksson <ulric@siag.nu>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -87,6 +87,7 @@ typedef struct {
 	unsigned char *downb, *downbptr, *upb, *upbptr;
 	int downn, upn;
 	int clt;
+	struct sockaddr_in *addr_backend, *addr_listener, *addr_client;		/* for UDP only */
 	int index;		/* server index */
 #ifdef HAVE_KQUEUE
 	int i;			/* connection index */
@@ -161,6 +162,10 @@ static int use_poll = 0;
 static int use_kqueue = 0;
 static int http = 0;
 static int client_acl, control_acl;
+static int udp = 0;
+static int udpused = 0;
+static int protoid = SOCK_STREAM;
+//static int connection_ctrl = 0;
 
 static int port;
 
@@ -179,9 +184,13 @@ static char *e_server = NULL;
 static char *a_server = NULL;
 static char *jail = NULL;
 static char *user = NULL;
+static char *proto = "tcp";
 
 static struct sigaction alrmaction, hupaction, termaction, usr1action;
 
+static void close_conn(int);
+static void no_response(void);
+
 static void debug(char *fmt, ...)
 {
 	time_t now;
@@ -330,9 +339,11 @@ static int ssl_init(void)
 	SSL_load_error_strings();
 	SSLeay_add_ssl_algorithms();
 	switch (ssl_protocol) {
+#if 0
 	case SRV_SSL_V2:
 		ssl_context = SSL_CTX_new(SSLv2_method());
 		break;
+#endif
 	case SRV_SSL_V3:
 		ssl_context = SSL_CTX_new(SSLv3_method());
 		break;
@@ -411,6 +422,9 @@ static int ssl_init(void)
 	SSL_CTX_set_client_CA_list(ssl_context,
 			SSL_load_client_CA_file(certfile));
 
+	/* permit large writes to be split up in several records */
+	SSL_CTX_set_mode(ssl_context, SSL_MODE_ENABLE_PARTIAL_WRITE);
+
 	return 0;
 }
 
@@ -749,9 +763,9 @@ static void dump(unsigned char *p, int n)
 }
 
 
-static int getport(char *p)
+static int getport(char *p, char *proto)
 {
-	struct servent *s = getservbyname(p, "tcp");
+	struct servent *s = getservbyname(p, proto);
 	if (s == NULL) {
 		return atoi(p);
 	} else {
@@ -772,14 +786,14 @@ static void setipaddress(struct in_addr *a, char *p)
 }
 
 static void setaddress(struct in_addr *a, int *port, char *s,
-		int dp, int *maxc, int *hard, int *weight, int *prio)
+		int dp, int *maxc, int *hard, int *weight, int *prio, char *proto)
 {
 	struct hostent *h;
 	char address[1024], pno[100];
 	int n = sscanf(s, "%999[^:]:%99[^:]:%d:%d:%d:%d",
 		address, pno, maxc, hard, weight, prio);
 
-	if (n > 1) *port = getport(pno);
+	if (n > 1) *port = getport(pno, proto);
 	else *port = dp;
 	if (n < 3) *maxc = 0;
 	if (n < 4) *hard = 0;
@@ -787,8 +801,8 @@ static void setaddress(struct in_addr *a, int *port, char *s,
 	if (n < 6) *prio = 0;
 
 	if (debuglevel)
-		debug("n = %d, address = %s, pno = %d, maxc1 = %d, hard = %d, weight = %d, prio = %d",
-			n, address, *port, *maxc, *hard, *weight, *prio);
+		debug("n = %d, address = %s, pno = %d, maxc1 = %d, hard = %d, weight = %d, prio = %d, proto = %s ",
+			n, address, *port, *maxc, *hard, *weight, *prio, proto);
 
 	if (!(h = gethostbyname(address))) {
 		if ((a->s_addr = inet_addr(address)) == -1) {
@@ -902,13 +916,17 @@ static int copy_up(int i)
 
 	unsigned char b[BUFFER_MAX];
 
+	size_t size = sizeof(struct sockaddr);
+
 #ifdef HAVE_SSL
 	SSL *ssl = conns[i].ssl;
 
 	if (ssl) {
 		rc = SSL_read(ssl, b, BUFFER_MAX);
+		if (debuglevel) debug("SSL_read returns %d\n", rc);
 		if (rc < 0) {
 			int err = SSL_get_error(ssl, rc);
+			if (debuglevel) debug("SSL_read returns %d (SSL error %d)\n", rc, err);
 			if (err == SSL_ERROR_WANT_READ ||
 			    err == SSL_ERROR_WANT_WRITE) {
 				return 0;
@@ -918,7 +936,13 @@ static int copy_up(int i)
 		rc = read(from, b, BUFFER_MAX);
 	}
 #else
-	rc = read(from, b, BUFFER_MAX);
+	if (!udp)
+		rc = read(from, b, BUFFER_MAX);
+	else {
+		rc = recvfrom (from, b, BUFFER_MAX, 0,
+                              (struct sockaddr *) conns[i].addr_listener, &size);
+
+	}
 #endif  /* HAVE_SSL */
 
 	if (debuglevel > 1) debug("copy_up(%d) %d bytes", i, rc);
@@ -942,7 +966,14 @@ static int copy_up(int i)
 		}
 
 		if (delayed_forward) n = 0;
-		else n = write(to, b, rc);	/* no ssl here */
+		else {
+			if (!udp)
+				n = write(to, b, rc);	/* no ssl here */
+			else {
+				n = sendto(to, b, rc, 0,
+					(struct sockaddr *) conns[i].addr_backend, size);
+			}
+		}
 
 		if (n < 0) {
 			if (!nblock || errno != EAGAIN) return -1;
@@ -975,7 +1006,13 @@ static int copy_down(int i)
 
 	unsigned char b[BUFFER_MAX];
 
-	rc = read(from, b, BUFFER_MAX);	/* no ssl here */
+	size_t size = sizeof(struct sockaddr);
+
+	if (!udp)
+		rc = read(from, b, BUFFER_MAX);	/* no ssl here */
+	else
+		rc = recvfrom (from, b, BUFFER_MAX, 0,
+                              (struct sockaddr *) conns[i].addr_backend, &size);
 
 	if (debuglevel > 1) debug("copy_down(%d) %d bytes", i, rc);
 	if (debuglevel > 2) dump(b, rc);
@@ -990,9 +1027,15 @@ static int copy_down(int i)
 		} else {
 #ifdef HAVE_SSL
 			if (ssl) {
-				n = SSL_write(ssl, b, rc);
+				/* can't write more than 32000 bytes at a time */
+				int ssl_rc;
+				if (rc > 32000) ssl_rc = 32000;
+				else ssl_rc = rc;
+				n = SSL_write(ssl, b, ssl_rc);
+				if (debuglevel) debug("SSL_write returns %d\n", n);
 				if (n < 0) {
 					int err = SSL_get_error(ssl, n);
+					if (debuglevel) debug("SSL error %d\n", err);
 					if (err == SSL_ERROR_WANT_READ ||
 					    err == SSL_ERROR_WANT_WRITE) {
 						return 0;
@@ -1002,7 +1045,12 @@ static int copy_down(int i)
 				n = write(to, b, rc);
 			}
 #else
-			n = write(to, b, rc);
+			if (!udp)
+				n = write(to, b, rc);
+			else {
+				n = sendto(to, b, rc, 0,
+					(struct sockaddr *) conns[i].addr_listener, size);
+			}
 #endif
 		}
 
@@ -1028,12 +1076,27 @@ static int copy_down(int i)
 static void alarm_handler(int dummy)
 {
 	if (debuglevel) debug("alarm_handler(%d)", dummy);
+	if (udp) no_response();
+}
+
+static void no_response(){		/* UDP packet is not responding */
+	int iserv = conns[0].index;
+	if (udpused == 2){
+		if (servers[iserv].status == 0){
+			if (debuglevel > 1) debug("No response UDP Server %d failed, \
+						retry in %d sec", iserv, blacklist_time);
+			servers[iserv].status = (int)time(NULL);
+		}
+	}
+	udpused=3;
 }
 
 #ifdef HAVE_SSL
-static void store_conn(int downfd, SSL *ssl, int upfd, int clt, int index)
+static void store_conn(int downfd, SSL *ssl, int upfd, int clt, int index,
+			struct sockaddr_in *serv_addr, struct sockaddr_in *cli_addr, struct sockaddr_in *addr_client)
 #else
-static void store_conn(int downfd, int upfd, int clt, int index)
+static void store_conn(int downfd, int upfd, int clt, int index,
+			struct sockaddr_in *serv_addr, struct sockaddr_in *cli_addr, struct sockaddr_in *addr_client)
 #endif
 {
 	int i, fl;
@@ -1067,6 +1130,11 @@ static void store_conn(int downfd, int upfd, int clt, int index)
 #endif
 		conns[i].clt = clt;
 		conns[i].index = index;
+		if (udp){
+			conns[i].addr_backend = serv_addr;
+			conns[i].addr_listener = cli_addr;
+			conns[i].addr_client = addr_client;
+		}
 #ifdef HAVE_KQUEUE
 		/* kqueue can store a pointer to arbitrary data with each
 		 * event, but it must be a pointer. We store a pointer to
@@ -1090,9 +1158,17 @@ static void store_conn(int downfd, int upfd, int clt, int index)
 static void close_conn(int i)
 {
 	int index = conns[i].index;
+//	if (connection_ctrl == 0) servers[index].c -= 1;
 	servers[index].c -= 1;
-	if (conns[i].upfd > 0) close(conns[i].upfd);
-	if (conns[i].downfd > 0) close(conns[i].downfd);
+	if (servers[index].c < 0) servers[index].c = 0;
+
+	if (conns[i].upfd > 0) {
+		close(conns[i].upfd);
+	}
+	if (conns[i].downfd > 0) {
+		if (!udp)
+			close(conns[i].downfd);
+	}
 	conns[i].upfd = conns[i].downfd = -1;
 	if (conns[i].downn) {
 		free(conns[i].downb);
@@ -1108,7 +1184,9 @@ static void close_conn(int i)
 		conns[i].ssl = 0;
 	}
 #endif
+//	connection_ctrl = 0;
 	connections_used--;
+	if (connections_used < 0) connections_used = 0;
 	if (debuglevel) debug("close_conn: connections_used = %d",
 				connections_used);
 }
@@ -1129,6 +1207,7 @@ static void usage(void)
 	       "  -C port   control port\n"
 	       "  -T sec    tracking time in seconds (0 = forever) [%d]\n"
 	       "  -H	add X-Forwarded-For header in http requests\n"
+	       "  -U	use udp protocol support\n"
 	       "  -P	use poll() rather than select()\n"
 	       "  -W    use weight for server selection\n"
 	       "  -X	enable 'exit' command for control port\n"
@@ -1218,7 +1297,7 @@ static void init(int argc, char **argv)
 		setaddress(&servers[server].addr, &servers[server].port,
 			argv[i], port,
 			&servers[server].maxc, &servers[server].hard,
-			&servers[server].weight, &servers[server].prio);
+			&servers[server].weight, &servers[server].prio, proto);
 		servers[server].sx = 0;
 		servers[server].rx = 0;
 
@@ -1231,7 +1310,7 @@ static void init(int argc, char **argv)
 		setaddress(&servers[server].addr, &servers[server].port,
 			"0.0.0.0", 0,
 			&servers[server].maxc, &servers[server].hard,
-			&servers[server].weight, &servers[server].prio);
+			&servers[server].weight, &servers[server].prio, proto);
 		servers[server].sx = 0;
 		servers[server].rx = 0;
 
@@ -1245,7 +1324,7 @@ static void init(int argc, char **argv)
 		setaddress(&servers[server].addr, &servers[server].port,
 			   e_server, port,
 			   &servers[server].maxc, &servers[server].hard,
-			   &servers[server].weight, &servers[server].prio);
+			   &servers[server].weight, &servers[server].prio, proto);
 		servers[server].sx = 0;
 		servers[server].rx = 0;
 		server++;
@@ -1258,7 +1337,7 @@ static void init(int argc, char **argv)
 		setaddress(&servers[server].addr, &servers[server].port,
 			   a_server, port,
 			   &servers[server].maxc, &servers[server].hard,
-			   &servers[server].weight, &servers[server].prio);
+			   &servers[server].weight, &servers[server].prio, proto);
 		servers[server].sx = 0;
 		servers[server].rx = 0;
 		server++;
@@ -1292,12 +1371,14 @@ static void init(int argc, char **argv)
 
 /* return upstream file descriptor */
 /* sticky = 1 if this client has used the server before */
-static int try_server(int index, int sticky, unsigned int cli_addr)
+static int try_server(int index, int sticky, unsigned int cli_addr,
+				struct sockaddr_in *addr, struct sockaddr_in *addr_client)
 {
-	struct sockaddr_in serv_addr;
 	int upfd;
-	int n;
+	int n = 0;
 	int now = (int)time(NULL);
+	int optval = 1;
+
 	if (debuglevel) debug("Trying server %d at time %d", index, now);
 	if (servers[index].port == 0) {
 		if (debuglevel) debug("No port for you!");
@@ -1320,15 +1401,18 @@ static int try_server(int index, int sticky, unsigned int cli_addr)
 		if (debuglevel) debug("try_server: denied by acl");
 		return -1;
 	}
-	upfd = socket(AF_INET, SOCK_STREAM, 0);
+
+	upfd = socket(AF_INET, protoid, 0);
 	if (upfd < 0) error("Error opening socket: %d", errno);
-	memset(&serv_addr, 0, sizeof serv_addr);
-	serv_addr.sin_family = AF_INET;
-	serv_addr.sin_addr.s_addr = servers[index].addr.s_addr;
-	serv_addr.sin_port = htons(servers[index].port);
+	memset(addr, 0, sizeof *addr);
+	addr->sin_family = AF_INET;
+	addr->sin_addr.s_addr = servers[index].addr.s_addr;
+	addr->sin_port = htons(servers[index].port);
+	setsockopt(upfd, SOL_SOCKET, SO_KEEPALIVE, &optval, sizeof optval);
+
 	sigaction(SIGALRM, &alrmaction, NULL);
 	alarm(timeout);
-	n = connect(upfd, (struct sockaddr *)&serv_addr, sizeof serv_addr);
+	n = connect(upfd, (struct sockaddr *) addr, sizeof *addr);
 	alarm(0);	/* cancel scheduled timeout, if there is one */
 	if (n == -1) {
 		if (servers[index].status == 0)
@@ -1339,7 +1423,7 @@ static int try_server(int index, int sticky, unsigned int cli_addr)
 	}
 	if (servers[index].status) {
 		servers[index].status=0;
-		debug("Server %d ok", index);
+		if (debuglevel) debug("Server %d ok", index);
 	}
 	if (debuglevel) debug("Successful connect to server %d", index);
 	return upfd;
@@ -1393,6 +1477,7 @@ static void write_cfg(char *p)
 	strftime(nowstr, sizeof(nowstr), "%Y-%m-%d %H:%M:%S", nowtm);
 	fprintf(fp, "# Generated by pen %s\n", nowstr);
 	fprintf(fp, "# pen");
+	if (udp) fprintf(fp, " -U");
 	if (foreground) fprintf(fp, " -f");
 	if (exit_enabled) fprintf(fp, " -X");
 	if (servers_max != SERVERS_MAX)
@@ -1825,6 +1910,7 @@ static void add_client(int downfd, struct sockaddr_in *cli_addr)
 {
 	int upfd = -1, clino = -1, index = -1, n, pd;
 	unsigned int cli = cli_addr->sin_addr.s_addr;
+	struct sockaddr_in serv_addr, addr_client;
 #ifdef HAVE_SSL
 	SSL *ssl = NULL;
 
@@ -1846,21 +1932,20 @@ static void add_client(int downfd, struct sockaddr_in *cli_addr)
 	if (!pd) {
 		if (debuglevel) debug("add_client: denied by acl");
 		if (abuse_server != -1) {
-			upfd = try_server(abuse_server, 0, cli);
+			upfd = try_server(abuse_server, 0, cli, &serv_addr, &addr_client);
 			if (upfd != -1) goto Success;
 		}
 		goto Failure;
 	}
 
-	if (roundrobin) {
-		if (debuglevel) debug("Bypassing client tracking");
-	} else {
+	if (!roundrobin) {
+		// Load balancing with memory == No roundrobin
 		clino = lookup_client(cli_addr->sin_addr);
 		if (debuglevel) debug("lookup_client returns %d", clino);
 		if (clino != -1) {
 			index = clients[clino].cno;
 			if (index != emerg_server && index != abuse_server) {
-				upfd = try_server(index, 1, cli);
+				upfd = try_server(index, 1, cli, &serv_addr, &addr_client);
 				if (upfd != -1) goto Success;
 			}
 		}
@@ -1868,20 +1953,20 @@ static void add_client(int downfd, struct sockaddr_in *cli_addr)
 	if (prio) {
 		index = server_by_prio();
 		if (index != -1) {
-			upfd = try_server(index, 0, cli);
+			upfd = try_server(index, 0, cli, &serv_addr, &addr_client);
 			if (upfd != -1) goto Success;
 		}
 	}
 	if (weight) {
 		index = server_by_weight();
 		if (index != -1) {
-			upfd = try_server(index, 0, cli);
+			upfd = try_server(index, 0, cli, &serv_addr, &addr_client);
 			if (upfd != -1) goto Success;
 		}
 	}
 	if (hash) {
 		index = cli % nservers;
-		upfd = try_server(index, 0, cli);
+		upfd = try_server(index, 0, cli, &serv_addr, &addr_client);
 		if (upfd != -1) goto Success;
 	}
 
@@ -1889,29 +1974,32 @@ static void add_client(int downfd, struct sockaddr_in *cli_addr)
 		index = current;
 		do {
 			index = (index + 1) % nservers;
-			if ((upfd = try_server(index, 0, cli)) != -1) goto Success;
+			if ((upfd = try_server(index, 0, cli, &serv_addr, &addr_client)) != -1) goto Success;
 		} while (index != current);
+		if (upfd == -1) goto Failure;
 	}
 	/* if we get here, we're dead */
 	if (emerg_server != -1) {
 		if (!emergency) 
 			debug("Using emergency server");
 		emergency=1;
-		if ((upfd = try_server(emerg_server, 0, cli)) != -1) goto Success2;
+		if ((upfd = try_server(emerg_server, 0, cli, &serv_addr, &addr_client)) != -1) goto Success2;
 	}
 	debug("Couldn't find a server for client");
 Failure:
+	if (!udp){
 	if (downfd != -1) close(downfd);
 	if (upfd != -1) close(upfd);
+	}
 	return;
 Success:
 	emergency=0;
 Success2:
 	n = store_client(clino, cli_addr->sin_addr, index);
 #ifdef HAVE_SSL
-	store_conn(downfd, ssl, upfd, n, index);
+	store_conn(downfd, ssl, upfd, n, index, &serv_addr, cli_addr, &addr_client);
 #else
-	store_conn(downfd, upfd, n, index);
+	store_conn(downfd, upfd, n, index, &serv_addr, cli_addr, &addr_client);
 #endif
 	return;
 }
@@ -1923,6 +2011,7 @@ static int open_listener(char *a)
 	char b[1024], *p;
 	int one = 1;
 	int fl;
+	int optval = 1;
 
 	memset(&serv_addr, 0, sizeof serv_addr);
 	serv_addr.sin_family = AF_INET;
@@ -1932,23 +2021,25 @@ static int open_listener(char *a)
 		b[sizeof b-1] = '\0';
 		p = strchr(b, ':');
 		*p = '\0';
-		port = getport(p+1);
+		port = getport(p+1, proto);
 		setipaddress(&serv_addr.sin_addr, b);
 		snprintf(b, (sizeof(b) - 1), "%s", inet_ntoa(serv_addr.sin_addr));
 	} else {
-		port = getport(a);
+		port = getport(a, proto);
 		serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
 		sprintf(b, "0.0.0.0");
 	}
 	serv_addr.sin_port = htons(port);
 
-	if ((listenfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
+	if ((listenfd = socket(AF_INET, protoid, 0)) < 0) {
 		error("can't open stream socket");
 	}
 
 	if (debuglevel) debug("local address=[%s:%d]", b, port);
 
 	setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR, (char *)&one, sizeof one);
+	setsockopt(listenfd, SOL_SOCKET, SO_KEEPALIVE, &optval, sizeof optval);
+
 	if (bind(listenfd, (struct sockaddr *) &serv_addr,
 		 sizeof serv_addr) < 0) {
 		error("can't bind local address");
@@ -1968,13 +2059,18 @@ static int flush_down(int i)
 {
 	int n;
 
+       struct sockaddr name;
+       size_t size = sizeof(struct sockaddr);
+
 #ifdef HAVE_SSL
 	SSL *ssl = conns[i].ssl;
 
 	if (ssl) {
 		n = SSL_write(ssl, conns[i].downbptr, conns[i].downn);
+		if (debuglevel) debug("SSL_write returns %d\n", n);
 		if (n < 0) {
 			int err = SSL_get_error(ssl, n);
+			if (debuglevel) debug("SSL_write returns %d (SSL error %d)", n, err);
 			if (err == SSL_ERROR_WANT_READ ||
 			    err == SSL_ERROR_WANT_WRITE) {
 				return 0;
@@ -1984,7 +2080,11 @@ static int flush_down(int i)
 		n = write(conns[i].downfd, conns[i].downbptr, conns[i].downn);
 	}
 #else
-	n = write(conns[i].downfd, conns[i].downbptr, conns[i].downn);
+	if (!udp)
+		n = write(conns[i].downfd, conns[i].downbptr, conns[i].downn);
+	else
+		n = sendto(conns[i].downfd, conns[i].downbptr, conns[i].downn, 0,
+			(struct sockaddr *) &name, size);
 #endif  /* HAVE_SSL */
 
 	if (debuglevel > 1) debug("flush_down(%d) %d bytes", i, n);
@@ -2002,7 +2102,16 @@ static int flush_down(int i)
 /* This only talks upstream and does not need ssl */
 static int flush_up(int i)
 {
-	int n = write(conns[i].upfd, conns[i].upbptr, conns[i].upn);
+	int n;
+
+       struct sockaddr name;
+       size_t size = sizeof(struct sockaddr);
+
+	if (!udp)
+		n = write(conns[i].upfd, conns[i].upbptr, conns[i].upn);
+	else
+		n = sendto(conns[i].upfd, conns[i].upbptr, conns[i].upn, 0,
+			(struct sockaddr *) &name, size);
 
 	if (debuglevel > 1) debug("flush_up(%d) %d bytes", i, n);
 	if (n > 0) {
@@ -2018,10 +2127,9 @@ static int flush_up(int i)
 static void mainloop_select(void)
 {
 	int downfd;
-	struct sockaddr_in cli_addr;
 	socklen_t clilen;
 	fd_set w_read, w_write, w_error;
-	int i, w_max;
+	int i, w_max, imax;
 	usr1action.sa_handler = stats;
 	sigemptyset(&usr1action.sa_mask);
 	usr1action.sa_flags = 0;
@@ -2073,8 +2181,10 @@ static void mainloop_select(void)
 					connections_used,
 					connections_max);
 		if (connections_used < connections_max) {
-			FD_SET(listenfd, &w_read);  /* new connections */
-			w_max = listenfd+1;
+			if (!udp || udpused == 0){
+				FD_SET(listenfd, &w_read);  /* new connections */
+				w_max = listenfd+1;
+			}
 		} else {
 			if (debuglevel) debug("Not listening");
 		}
@@ -2084,7 +2194,12 @@ static void mainloop_select(void)
 		}
 
 		/* add sockets from open connections */
-		for (i = 0; i < connections_max; i++) {
+		if (udp) /* udp has a secuential proccess */
+			imax = 1;
+		else
+			imax = connections_max;
+
+		for (i = 0; i < imax; i++) {
 			if (conns[i].downfd == -1) continue;
 			if (conns[i].upn == 0) {
 				FD_SET(conns[i].downfd, &w_read);
@@ -2118,23 +2233,33 @@ static void mainloop_select(void)
 		}
 		if (n <= 0) continue;
 
+	if (!udp || (udp && (udpused == 0))){
 		if (FD_ISSET(listenfd, &w_read)) {
+			struct sockaddr_in cli_addr;
 			clilen = sizeof cli_addr;
-			downfd = accept(listenfd,
-				(struct sockaddr *) &cli_addr, &clilen);
-			if (downfd < 0) {
-				if (debuglevel) perror("accept");
-				continue;
-			}
-			if (clilen == 0) {
-				if (debuglevel) perror("clilen");
-				continue;
+			if (!udp) {
+				downfd = accept(listenfd,
+					(struct sockaddr *) &cli_addr, &clilen);
+				if (downfd < 0) {
+					if (debuglevel) perror("accept");
+					continue;
+				}
+				if (clilen == 0) {
+					if (debuglevel) perror("clilen");
+					continue;
+				}
+			} else {
+				downfd = listenfd;
+				udpused = 1;
 			}
 			add_client(downfd, &cli_addr);
 		}
+	}
 
 		/* check control port */
 		if (ctrlfd != -1 && FD_ISSET(ctrlfd, &w_read)) {
+//			connection_ctrl = 1;
+			struct sockaddr_in cli_addr;
 			clilen = sizeof cli_addr;
 			downfd = accept(ctrlfd,
 				(struct sockaddr *) &cli_addr, &clilen);
@@ -2150,6 +2275,32 @@ static void mainloop_select(void)
 		}
 
 		/* check sockets from open connections */
+		if (udp){
+			i=0;
+			if (conns[i].downfd == -1 || conns[i].upfd == -1 || udpused==3){
+				alarm(0);
+				close_conn(i);
+				udpused=0;
+			} else {
+			if (udpused!=2){
+				sigaction(SIGALRM, &alrmaction, NULL);
+				alarm(timeout);
+				if (FD_ISSET(conns[i].downfd, &w_read)) {
+					udpused = 2;
+					if (copy_up(i) < 0) {
+						udpused = 0;
+						close_conn(i);
+					}
+				}
+			}
+			if (FD_ISSET(conns[i].upfd, &w_read)) {
+				copy_down(i);
+				close_conn(i);
+				alarm(0);
+				udpused=0;
+			}
+			}
+		} else {		/* TCP */
 		for (i = 0; i < connections_max; i++) {
 			if (conns[i].downfd == -1) continue;
 			if (FD_ISSET(conns[i].downfd, &w_read)) {
@@ -2177,6 +2328,7 @@ static void mainloop_select(void)
 				}
 			}
 		}
+		}
 	}
 }
 
@@ -2563,9 +2715,9 @@ static int options(int argc, char **argv)
 	char b[1024];
 
 #ifdef HAVE_SSL
-	char *opt = "B:C:F:S:T:b:c:e:j:l:o:p:t:u:w:x:DHPQWXadfhnrsE:K:G:A:ZRL:";
+	char *opt = "B:C:F:S:T:b:c:e:j:l:o:p:t:u:w:x:DHPQWXUadfhnrsE:K:G:A:ZRL:";
 #else
-	char *opt = "B:C:F:S:T:b:c:e:j:l:o:p:t:u:w:x:DHPQWXadfhnrs";
+	char *opt = "B:C:F:S:T:b:c:e:j:l:o:p:t:u:w:x:DHPQWXUadfhnrs";
 #endif
 
 	while ((c = getopt(argc, argv, opt)) != -1) {
@@ -2597,6 +2749,9 @@ static int options(int argc, char **argv)
 		case 'T':
 			tracking_time = atoi(optarg);
 			break;
+		case 'U':
+			udp = 1;
+			break;
 		case 'W':
 			weight = 1;
 			break;
@@ -2739,9 +2894,16 @@ int main(int argc, char **argv)
 #endif
 
 	/* we must open listeners before dropping privileges */
+	/* Control port */
 	if (ctrlport) ctrlfd = open_listener(ctrlport);
 	else ctrlfd = -1;
 
+	/* Balancing port */
+	if (udp) {
+		protoid = SOCK_DGRAM;
+		proto = "udp";
+	}
+
 	listenport = argv[0];
 	listenfd = open_listener(listenport);
 	init(argc, argv);
